@@ -18,8 +18,6 @@
  * Or go to http://www.gnu.org/copyleft/lgpl.html
  */
 
-#define _CRT_SECURE_NO_DEPRECATE // get rid of sprintf security warnings on VS2005
-
 #include "config.h"
 
 #include <math.h>
@@ -38,16 +36,6 @@
 #include "alAuxEffectSlot.h"
 #include "alu.h"
 #include "bs2b.h"
-#include "alReverb.h"
-
-#if defined (HAVE_FLOAT_H)
-#include <float.h>
-#endif
-
-#ifndef M_PI
-#define M_PI           3.14159265358979323846  /* pi */
-#define M_PI_2         1.57079632679489661923  /* pi/2 */
-#endif
 
 #if defined(HAVE_STDINT_H)
 #include <stdint.h>
@@ -60,39 +48,6 @@ typedef long ALint64;
 typedef long long ALint64;
 #endif
 
-#ifdef HAVE_SQRTF
-#define aluSqrt(x) ((ALfloat)sqrtf((float)(x)))
-#else
-#define aluSqrt(x) ((ALfloat)sqrt((double)(x)))
-#endif
-
-#ifdef HAVE_ACOSF
-#define aluAcos(x) ((ALfloat)acosf((float)(x)))
-#else
-#define aluAcos(x) ((ALfloat)acos((double)(x)))
-#endif
-
-#ifdef HAVE_ATANF
-#define aluAtan(x) ((ALfloat)atanf((float)(x)))
-#else
-#define aluAtan(x) ((ALfloat)atan((double)(x)))
-#endif
-
-#ifdef HAVE_FABSF
-#define aluFabs(x) ((ALfloat)fabsf((float)(x)))
-#else
-#define aluFabs(x) ((ALfloat)fabs((double)(x)))
-#endif
-
-// fixes for mingw32.
-#if defined(max) && !defined(__max)
-#define __max max
-#endif
-#if defined(min) && !defined(__min)
-#define __min min
-#endif
-
-#define BUFFERSIZE 24000
 #define FRACTIONBITS 14
 #define FRACTIONMASK ((1L<<FRACTIONBITS)-1)
 #define MAX_PITCH 65536
@@ -103,117 +58,6 @@ typedef long long ALint64;
 
 ALboolean DuplicateStereo = AL_FALSE;
 
-/* NOTE: The AL_FORMAT_REAR* enums aren't handled here be cause they're
- *       converted to AL_FORMAT_QUAD* when loaded */
-__inline ALuint aluBytesFromFormat(ALenum format)
-{
-    switch(format)
-    {
-        case AL_FORMAT_MONO8:
-        case AL_FORMAT_STEREO8:
-        case AL_FORMAT_QUAD8_LOKI:
-        case AL_FORMAT_QUAD8:
-        case AL_FORMAT_51CHN8:
-        case AL_FORMAT_61CHN8:
-        case AL_FORMAT_71CHN8:
-            return 1;
-
-        case AL_FORMAT_MONO16:
-        case AL_FORMAT_STEREO16:
-        case AL_FORMAT_QUAD16_LOKI:
-        case AL_FORMAT_QUAD16:
-        case AL_FORMAT_51CHN16:
-        case AL_FORMAT_61CHN16:
-        case AL_FORMAT_71CHN16:
-            return 2;
-
-        case AL_FORMAT_MONO_FLOAT32:
-        case AL_FORMAT_STEREO_FLOAT32:
-        case AL_FORMAT_QUAD32:
-        case AL_FORMAT_51CHN32:
-        case AL_FORMAT_61CHN32:
-        case AL_FORMAT_71CHN32:
-            return 4;
-
-        default:
-            return 0;
-    }
-}
-
-__inline ALuint aluChannelsFromFormat(ALenum format)
-{
-    switch(format)
-    {
-        case AL_FORMAT_MONO8:
-        case AL_FORMAT_MONO16:
-        case AL_FORMAT_MONO_FLOAT32:
-            return 1;
-
-        case AL_FORMAT_STEREO8:
-        case AL_FORMAT_STEREO16:
-        case AL_FORMAT_STEREO_FLOAT32:
-            return 2;
-
-        case AL_FORMAT_QUAD8_LOKI:
-        case AL_FORMAT_QUAD16_LOKI:
-        case AL_FORMAT_QUAD8:
-        case AL_FORMAT_QUAD16:
-        case AL_FORMAT_QUAD32:
-            return 4;
-
-        case AL_FORMAT_51CHN8:
-        case AL_FORMAT_51CHN16:
-        case AL_FORMAT_51CHN32:
-            return 6;
-
-        case AL_FORMAT_61CHN8:
-        case AL_FORMAT_61CHN16:
-        case AL_FORMAT_61CHN32:
-            return 7;
-
-        case AL_FORMAT_71CHN8:
-        case AL_FORMAT_71CHN16:
-        case AL_FORMAT_71CHN32:
-            return 8;
-
-        default:
-            return 0;
-    }
-}
-
-
-static __inline ALfloat lpFilter(FILTER *iir, ALfloat input)
-{
-    ALfloat *history = iir->history;
-    ALfloat a = iir->coeff;
-    ALfloat output = input;
-
-    output = output + (history[0]-output)*a;
-    history[0] = output;
-    output = output + (history[1]-output)*a;
-    history[1] = output;
-    output = output + (history[2]-output)*a;
-    history[2] = output;
-    output = output + (history[3]-output)*a;
-    history[3] = output;
-
-    return output;
-}
-
-static __inline ALfloat lpFilterMC(FILTER *iir, ALuint chan, ALfloat input)
-{
-    ALfloat *history = &iir->history[chan*2];
-    ALfloat a = iir->coeff;
-    ALfloat output = input;
-
-    output = output + (history[0]-output)*a;
-    history[0] = output;
-    output = output + (history[1]-output)*a;
-    history[1] = output;
-
-    return output;
-}
-
 
 static __inline ALshort aluF2S(ALfloat Value)
 {
@@ -560,27 +404,29 @@ static ALvoid CalcSourceParams(const ALCcontext *ALContext,
                                ALfloat *pitch, ALfloat *drygainhf,
                                ALfloat *wetgainhf)
 {
-    ALfloat InnerAngle,OuterAngle,Angle,Distance,DryMix,WetMix=0.0f;
+    ALfloat InnerAngle,OuterAngle,Angle,Distance,DryMix;
     ALfloat Direction[3],Position[3],SourceToListener[3];
     ALfloat MinVolume,MaxVolume,MinDist,MaxDist,Rolloff,OuterGainHF;
-    ALfloat ConeVolume,SourceVolume,ListenerGain;
+    ALfloat ConeVolume,ConeHF,SourceVolume,ListenerGain;
     ALfloat U[3],V[3],N[3];
     ALfloat DopplerFactor, DopplerVelocity, flSpeedOfSound, flMaxVelocity;
     ALfloat Matrix[3][3];
     ALfloat flAttenuation;
-    ALfloat RoomAttenuation;
+    ALfloat RoomAttenuation[MAX_SENDS];
     ALfloat MetersPerUnit;
-    ALfloat RoomRolloff;
+    ALfloat RoomRolloff[MAX_SENDS];
     ALfloat DryGainHF = 1.0f;
-    ALfloat WetGainHF = 1.0f;
     ALfloat DirGain, AmbientGain;
+    ALfloat length;
     const ALfloat *SpeakerGain;
-    ALint pos, s;
+    ALint NumSends;
+    ALint pos, s, i;
 
     //Get context properties
     DopplerFactor   = ALContext->DopplerFactor * ALSource->DopplerFactor;
     DopplerVelocity = ALContext->DopplerVelocity;
     flSpeedOfSound  = ALContext->flSpeedOfSound;
+    NumSends        = ALContext->NumSends;
 
     //Get listener properties
     ListenerGain = ALContext->Listener.Gain;
@@ -598,7 +444,6 @@ static ALvoid CalcSourceParams(const ALCcontext *ALContext,
     InnerAngle   = ALSource->flInnerAngle;
     OuterAngle   = ALSource->flOuterAngle;
     OuterGainHF  = ALSource->OuterGainHF;
-    RoomRolloff  = ALSource->RoomRolloffFactor;
 
     //Only apply 3D calculations for mono buffers
     if(isMono != AL_FALSE)
@@ -645,14 +490,17 @@ static ALvoid CalcSourceParams(const ALCcontext *ALContext,
         //2. Calculate distance attenuation
         Distance = aluSqrt(aluDotproduct(Position, Position));
 
-        if(ALSource->Send[0].Slot)
+        flAttenuation = 1.0f;
+        for(i = 0;i < MAX_SENDS;i++)
         {
-            if(ALSource->Send[0].Slot->effect.type == AL_EFFECT_REVERB)
-                RoomRolloff += ALSource->Send[0].Slot->effect.Reverb.RoomRolloffFactor;
+            RoomAttenuation[i] = 1.0f;
+
+            RoomRolloff[i] = ALSource->RoomRolloffFactor;
+            if(ALSource->Send[i].Slot &&
+               ALSource->Send[i].Slot->effect.type == AL_EFFECT_REVERB)
+                RoomRolloff[i] += ALSource->Send[i].Slot->effect.Reverb.RoomRolloffFactor;
         }
 
-        flAttenuation = 1.0f;
-        RoomAttenuation = 1.0f;
         switch (ALSource->DistanceModel)
         {
             case AL_INVERSE_DISTANCE_CLAMPED:
@@ -666,8 +514,11 @@ static ALvoid CalcSourceParams(const ALCcontext *ALContext,
                 {
                     if ((MinDist + (Rolloff * (Distance - MinDist))) > 0.0f)
                         flAttenuation = MinDist / (MinDist + (Rolloff * (Distance - MinDist)));
-                    if ((MinDist + (RoomRolloff * (Distance - MinDist))) > 0.0f)
-                        RoomAttenuation = MinDist / (MinDist + (RoomRolloff * (Distance - MinDist)));
+                    for(i = 0;i < NumSends;i++)
+                    {
+                        if ((MinDist + (RoomRolloff[i] * (Distance - MinDist))) > 0.0f)
+                            RoomAttenuation[i] = MinDist / (MinDist + (RoomRolloff[i] * (Distance - MinDist)));
+                    }
                 }
                 break;
 
@@ -682,7 +533,8 @@ static ALvoid CalcSourceParams(const ALCcontext *ALContext,
                 if (MaxDist != MinDist)
                 {
                     flAttenuation = 1.0f - (Rolloff*(Distance-MinDist)/(MaxDist - MinDist));
-                    RoomAttenuation = 1.0f - (RoomRolloff*(Distance-MinDist)/(MaxDist - MinDist));
+                    for(i = 0;i < NumSends;i++)
+                        RoomAttenuation[i] = 1.0f - (RoomRolloff[i]*(Distance-MinDist)/(MaxDist - MinDist));
                 }
                 break;
 
@@ -696,18 +548,30 @@ static ALvoid CalcSourceParams(const ALCcontext *ALContext,
                 if ((Distance > 0.0f) && (MinDist > 0.0f))
                 {
                     flAttenuation = (ALfloat)pow(Distance/MinDist, -Rolloff);
-                    RoomAttenuation = (ALfloat)pow(Distance/MinDist, -RoomRolloff);
+                    for(i = 0;i < NumSends;i++)
+                        RoomAttenuation[i] = (ALfloat)pow(Distance/MinDist, -RoomRolloff[i]);
                 }
                 break;
 
             case AL_NONE:
-                flAttenuation = 1.0f;
-                RoomAttenuation = 1.0f;
                 break;
         }
 
+        // Source Gain + Attenuation and clamp to Min/Max Gain
+        DryMix = SourceVolume * flAttenuation;
+        DryMix = __min(DryMix,MaxVolume);
+        DryMix = __max(DryMix,MinVolume);
+
+        for(i = 0;i < NumSends;i++)
+        {
+            ALfloat WetMix = SourceVolume * RoomAttenuation[i];
+            WetMix = __min(WetMix,MaxVolume);
+            wetsend[i] = __max(WetMix,MinVolume);
+            wetgainhf[i] = 1.0f;
+        }
+
         // Distance-based air absorption
-        if(ALSource->AirAbsorptionFactor > 0.0f && ALContext->DistanceModel != AL_NONE)
+        if(ALSource->AirAbsorptionFactor > 0.0f && ALSource->DistanceModel != AL_NONE)
         {
             ALfloat dist = Distance-MinDist;
             ALfloat absorb;
@@ -719,42 +583,33 @@ static ALvoid CalcSourceParams(const ALCcontext *ALContext,
             // Convert dB to linear gain before applying
             absorb = pow(10.0, absorb/20.0);
             DryGainHF *= absorb;
-            WetGainHF *= absorb;
+            for(i = 0;i < MAX_SENDS;i++)
+                wetgainhf[i] *= absorb;
         }
 
-        // Source Gain + Attenuation and clamp to Min/Max Gain
-        DryMix = SourceVolume * flAttenuation;
-        DryMix = __min(DryMix,MaxVolume);
-        DryMix = __max(DryMix,MinVolume);
-
-        WetMix = SourceVolume * RoomAttenuation;
-        WetMix = __min(WetMix,MaxVolume);
-        WetMix = __max(WetMix,MinVolume);
-
         //3. Apply directional soundcones
         Angle = aluAcos(aluDotproduct(Direction,SourceToListener)) * 180.0f/M_PI;
         if(Angle >= InnerAngle && Angle <= OuterAngle)
         {
             ALfloat scale = (Angle-InnerAngle) / (OuterAngle-InnerAngle);
             ConeVolume = (1.0f+(ALSource->flOuterGain-1.0f)*scale);
+            ConeHF = (1.0f+(OuterGainHF-1.0f)*scale);
             DryMix *= ConeVolume;
-            if(ALSource->WetGainAuto)
-                WetMix *= ConeVolume;
             if(ALSource->DryGainHFAuto)
-                DryGainHF *= (1.0f+(OuterGainHF-1.0f)*scale);
-            if(ALSource->WetGainHFAuto)
-                WetGainHF *= (1.0f+(OuterGainHF-1.0f)*scale);
+                DryGainHF *= ConeHF;
         }
         else if(Angle > OuterAngle)
         {
             ConeVolume = (1.0f+(ALSource->flOuterGain-1.0f));
+            ConeHF = (1.0f+(OuterGainHF-1.0f));
             DryMix *= ConeVolume;
-            if(ALSource->WetGainAuto)
-                WetMix *= ConeVolume;
             if(ALSource->DryGainHFAuto)
-                DryGainHF *= (1.0f+(OuterGainHF-1.0f));
-            if(ALSource->WetGainHFAuto)
-                WetGainHF *= (1.0f+(OuterGainHF-1.0f));
+                DryGainHF *= ConeHF;
+        }
+        else
+        {
+            ConeVolume = 1.0f;
+            ConeHF = 1.0f;
         }
 
         //4. Calculate Velocity
@@ -785,32 +640,49 @@ static ALvoid CalcSourceParams(const ALCcontext *ALContext,
         else
             pitch[0] = ALSource->flPitch;
 
-        if(ALSource->Send[0].Slot &&
-           ALSource->Send[0].Slot->effect.type != AL_EFFECT_NULL)
+        for(i = 0;i < NumSends;i++)
         {
-            if(ALSource->Send[0].Slot->AuxSendAuto)
+            if(ALSource->Send[i].Slot &&
+               ALSource->Send[i].Slot->effect.type != AL_EFFECT_NULL)
             {
-                // Apply minimal attenuation in place of missing statistical
-                // reverb model.
-                WetMix *= pow(DryMix, 1.0f / 2.0f);
+                if(ALSource->WetGainAuto)
+                    wetsend[i] *= ConeVolume;
+                if(ALSource->WetGainHFAuto)
+                    wetgainhf[i] *= ConeHF;
+
+                if(ALSource->Send[i].Slot->AuxSendAuto)
+                {
+                    // Apply minimal attenuation in place of missing
+                    // statistical reverb model.
+                    wetsend[i] *= pow(DryMix, 1.0f / 2.0f);
+                }
+                else
+                {
+                    // If the slot's auxilliary send auto is off, the data sent to the
+                    // effect slot is the same as the dry path, sans filter effects
+                    wetsend[i] = DryMix;
+                    wetgainhf[i] = DryGainHF;
+                }
+
+                switch(ALSource->Send[i].WetFilter.type)
+                {
+                    case AL_FILTER_LOWPASS:
+                        wetsend[i] *= ALSource->Send[i].WetFilter.Gain;
+                        wetgainhf[i] *= ALSource->Send[i].WetFilter.GainHF;
+                        break;
+                }
+                wetsend[i] *= ListenerGain;
             }
             else
             {
-                // If the slot's auxilliary send auto is off, the data sent to the
-                // effect slot is the same as the dry path, sans filter effects
-                WetMix = DryMix;
-                WetGainHF = DryGainHF;
+                wetsend[i] = 0.0f;
+                wetgainhf[i] = 1.0f;
             }
-
-            // Note that this is really applied by the effect slot. However,
-            // it's easier (more optimal) to handle it here.
-            if(ALSource->Send[0].Slot->effect.type == AL_EFFECT_REVERB)
-                WetGainHF *= ALSource->Send[0].Slot->effect.Reverb.GainHF;
         }
-        else
+        for(i = NumSends;i < MAX_SENDS;i++)
         {
-            WetMix = 0.0f;
-            WetGainHF = 1.0f;
+            wetsend[i] = 0.0f;
+            wetgainhf[i] = 1.0f;
         }
 
         //5. Apply filter gains and filters
@@ -821,20 +693,19 @@ static ALvoid CalcSourceParams(const ALCcontext *ALContext,
                 DryGainHF *= ALSource->DirectFilter.GainHF;
                 break;
         }
-
-        switch(ALSource->Send[0].WetFilter.type)
-        {
-            case AL_FILTER_LOWPASS:
-                WetMix *= ALSource->Send[0].WetFilter.Gain;
-                WetGainHF *= ALSource->Send[0].WetFilter.GainHF;
-                break;
-        }
-
         DryMix *= ListenerGain;
-        WetMix *= ListenerGain;
 
         // Use energy-preserving panning algorithm for multi-speaker playback
-        aluNormalize(Position);
+        length = aluSqrt(Position[0]*Position[0] + Position[1]*Position[1] +
+                         Position[2]*Position[2]);
+        length = __max(length, MinDist);
+        if(length > 0.0f)
+        {
+            ALfloat invlen = 1.0f/length;
+            Position[0] *= invlen;
+            Position[1] *= invlen;
+            Position[2] *= invlen;
+        }
 
         pos = aluCart2LUTpos(-Position[2], Position[0]);
         SpeakerGain = &ALContext->PanningLUT[OUTPUTCHANNELS * pos];
@@ -848,10 +719,7 @@ static ALvoid CalcSourceParams(const ALCcontext *ALContext,
             ALfloat gain = SpeakerGain[s]*DirGain + AmbientGain;
             drysend[s] = DryMix * gain;
         }
-        *wetsend = WetMix;
-
         *drygainhf = DryGainHF;
-        *wetgainhf = WetGainHF;
     }
     else
     {
@@ -879,10 +747,13 @@ static ALvoid CalcSourceParams(const ALCcontext *ALContext,
         drysend[FRONT_CENTER] = DryMix * ListenerGain;
         drysend[BACK_CENTER]  = DryMix * ListenerGain;
         drysend[LFE]          = DryMix * ListenerGain;
-        *wetsend              = 0.0f;
+        *drygainhf            = DryGainHF;
 
-        *drygainhf = DryGainHF;
-        *wetgainhf = WetGainHF;
+        for(i = 0;i < MAX_SENDS;i++)
+        {
+            wetsend[i] = 0.0f;
+            wetgainhf[i] = 1.0f;
+        }
     }
 }
 
@@ -894,17 +765,16 @@ static __inline ALshort lerp(ALshort val1, ALshort val2, ALint frac)
 ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum format)
 {
     static float DryBuffer[BUFFERSIZE][OUTPUTCHANNELS];
-    static float WetBuffer[BUFFERSIZE];
+    static float DummyBuffer[BUFFERSIZE];
+    ALfloat *WetBuffer[MAX_SENDS];
     ALfloat (*Matrix)[OUTPUTCHANNELS] = ALContext->ChannelMatrix;
-    ALfloat newDrySend[OUTPUTCHANNELS] = { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f };
-    ALfloat newWetSend = 0.0f;
+    ALfloat DrySend[OUTPUTCHANNELS] = { 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f };
+    ALfloat WetSend[MAX_SENDS];
     ALfloat DryGainHF = 0.0f;
-    ALfloat WetGainHF = 0.0f;
-    ALfloat *DrySend;
-    ALfloat *WetSend;
+    ALfloat WetGainHF[MAX_SENDS];
     ALuint rampLength;
     ALfloat dryGainStep[OUTPUTCHANNELS];
-    ALfloat wetGainStep;
+    ALfloat wetGainStep[MAX_SENDS];
     ALuint BlockAlign,BufferSize;
     ALuint DataSize=0,DataPosInt=0,DataPosFrac=0;
     ALuint Channels,Frequency,ulExtraSamples;
@@ -924,7 +794,7 @@ ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum forma
     ALbufferlistitem *BufferListItem;
     ALuint loop;
     ALint64 DataSize64,DataPos64;
-    FILTER *DryFilter, *WetFilter;
+    FILTER *DryFilter, *WetFilter[MAX_SENDS];
     int fpuState;
 
     SuspendContext(ALContext);
@@ -963,7 +833,6 @@ ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum forma
         rampLength = max(rampLength, SamplesToDo);
 
         //Clear mixing buffer
-        memset(WetBuffer, 0, SamplesToDo*sizeof(ALfloat));
         memset(DryBuffer, 0, SamplesToDo*OUTPUTCHANNELS*sizeof(ALfloat));
 
         //Actual mixing loop
@@ -996,14 +865,18 @@ ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum forma
 
                     //Get source info
                     DryFilter = &ALSource->iirFilter;
-                    WetFilter = &ALSource->Send[0].iirFilter;
-                    DrySend = ALSource->DryGains;
-                    WetSend = &ALSource->WetGain;
+                    for(i = 0;i < MAX_SENDS;i++)
+                    {
+                        WetFilter[i] = &ALSource->Send[i].iirFilter;
+                        WetBuffer[i] = (ALSource->Send[i].Slot ?
+                                        ALSource->Send[i].Slot->WetBuffer :
+                                        DummyBuffer);
+                    }
 
                     CalcSourceParams(ALContext, ALSource,
                                      (Channels==1) ? AL_TRUE : AL_FALSE,
-                                     newDrySend, &newWetSend, &Pitch,
-                                     &DryGainHF, &WetGainHF);
+                                     DrySend, WetSend, &Pitch,
+                                     &DryGainHF, WetGainHF);
                     Pitch = (Pitch*Frequency) / ALContext->Frequency;
 
                     if(Channels == 1)
@@ -1023,24 +896,30 @@ ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum forma
                             a = (1 - g*cw - aluSqrt(2*g*(1-cw) - g*g*(1 - cw*cw))) / (1 - g);
                         DryFilter->coeff = a;
 
-                        g = aluSqrt(__max(WetGainHF, 0.0001f));
-                        a = 0.0f;
-                        if(g < 0.9999f) // 1-epsilon
-                            a = (1 - g*cw - aluSqrt(2*g*(1-cw) - g*g*(1 - cw*cw))) / (1 - g);
-                        WetFilter->coeff = a;
+                        for(i = 0;i < MAX_SENDS;i++)
+                        {
+                            // The wet path uses two chained one-pole filters,
+                            // so take the base gain (square root of the
+                            // squared gain)
+                            g = __max(WetGainHF[i], 0.01f);
+                            a = 0.0f;
+                            if(g < 0.9999f) // 1-epsilon
+                                a = (1 - g*cw - aluSqrt(2*g*(1-cw) - g*g*(1 - cw*cw))) / (1 - g);
+                            WetFilter[i]->coeff = a;
+                        }
                     }
                     else
                     {
                         // Multi-channel sources use two chained one-pole
-                        // filters, so take the base gain (square root of the
-                        // squared gain)
+                        // filters
                         cw = cos(2.0*M_PI * LOWPASSFREQCUTOFF / ALContext->Frequency);
                         g = __max(DryGainHF, 0.01f);
                         a = 0.0f;
                         if(g < 0.9999f) // 1-epsilon
                             a = (1 - g*cw - aluSqrt(2*g*(1-cw) - g*g*(1 - cw*cw))) / (1 - g);
                         DryFilter->coeff = a;
-                        WetFilter->coeff = 0.0f;
+                        for(i = 0;i < MAX_SENDS;i++)
+                            WetFilter[i]->coeff = 0.0f;
 
                         if(DuplicateStereo && Channels == 2)
                         {
@@ -1062,18 +941,22 @@ ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum forma
                     if(ALSource->FirstStart && DataPosInt == 0 && DataPosFrac == 0)
                     {
                         for(i = 0;i < OUTPUTCHANNELS;i++)
-                        {
-                            DrySend[i] = newDrySend[i];
-                            dryGainStep[i] = 0;
-                        }
-                        *WetSend = newWetSend;
-                        wetGainStep = 0;
+                            dryGainStep[i] = 0.0f;
+                        for(i = 0;i < MAX_SENDS;i++)
+                            wetGainStep[i] = 0.0f;
                     }
                     else
                     {
                         for(i = 0;i < OUTPUTCHANNELS;i++)
-                            dryGainStep[i] = (newDrySend[i]-DrySend[i]) / rampLength;
-                        wetGainStep = (newWetSend-(*WetSend)) / rampLength;
+                        {
+                            dryGainStep[i] = (DrySend[i]-ALSource->DryGains[i]) / rampLength;
+                            DrySend[i] = ALSource->DryGains[i];
+                        }
+                        for(i = 0;i < MAX_SENDS;i++)
+                        {
+                            wetGainStep[i] = (WetSend[i]-ALSource->WetGains[i]) / rampLength;
+                            WetSend[i] = ALSource->WetGains[i];
+                        }
                     }
                     ALSource->FirstStart = AL_FALSE;
 
@@ -1135,13 +1018,14 @@ ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum forma
                         {
                             for(i = 0;i < OUTPUTCHANNELS;i++)
                                 DrySend[i] += dryGainStep[i];
-                            *WetSend += wetGainStep;
+                            for(i = 0;i < MAX_SENDS;i++)
+                                WetSend[i] += wetGainStep[i];
 
                             //First order interpolator
                             value = lerp(Data[k], Data[k+1], DataPosFrac);
 
                             //Direct path final mix buffer and panning
-                            outsamp = lpFilter(DryFilter, value);
+                            outsamp = lpFilter4P(DryFilter, 0, value);
                             DryBuffer[j][FRONT_LEFT]   += outsamp*DrySend[FRONT_LEFT];
                             DryBuffer[j][FRONT_RIGHT]  += outsamp*DrySend[FRONT_RIGHT];
                             DryBuffer[j][SIDE_LEFT]    += outsamp*DrySend[SIDE_LEFT];
@@ -1152,8 +1036,11 @@ ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum forma
                             DryBuffer[j][BACK_CENTER]  += outsamp*DrySend[BACK_CENTER];
 
                             //Room path final mix buffer and panning
-                            outsamp = lpFilter(WetFilter, value);
-                            WetBuffer[j] += outsamp*(*WetSend);
+                            for(i = 0;i < MAX_SENDS;i++)
+                            {
+                                outsamp = lpFilter2P(WetFilter[i], 0, value);
+                                WetBuffer[i][j] += outsamp*WetSend[i];
+                            }
 
                             DataPosFrac += increment;
                             k += DataPosFrac>>FRACTIONBITS;
@@ -1168,7 +1055,8 @@ ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum forma
                         };
 
 #define DO_MIX() do { \
-    *WetSend += wetGainStep*BufferSize; \
+    for(i = 0;i < MAX_SENDS;i++) \
+        WetSend[i] += wetGainStep[i]*BufferSize; \
     while(BufferSize--) \
     { \
         for(i = 0;i < OUTPUTCHANNELS;i++) \
@@ -1177,7 +1065,7 @@ ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum forma
         for(i = 0;i < Channels;i++) \
         { \
             value = lerp(Data[k*Channels + i], Data[(k+1)*Channels + i], DataPosFrac); \
-            values[i] = lpFilterMC(DryFilter, chans[i], value)*DrySend[chans[i]]; \
+            values[i] = lpFilter2P(DryFilter, chans[i]*2, value)*DrySend[chans[i]]; \
         } \
         for(out = 0;out < OUTPUTCHANNELS;out++) \
         { \
@@ -1240,9 +1128,10 @@ ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum forma
                     }
                     else /* Unknown? */
                     {
-                        *WetSend += wetGainStep*BufferSize;
                         for(i = 0;i < OUTPUTCHANNELS;i++)
                             DrySend[i] += dryGainStep[i]*BufferSize;
+                        for(i = 0;i < MAX_SENDS;i++)
+                            WetSend[i] += wetGainStep[i]*BufferSize;
                         while(BufferSize--)
                         {
                             DataPosFrac += increment;
@@ -1256,6 +1145,10 @@ ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum forma
                     //Update source info
                     ALSource->position = DataPosInt;
                     ALSource->position_fraction = DataPosFrac;
+                    for(i = 0;i < OUTPUTCHANNELS;i++)
+                        ALSource->DryGains[i] = DrySend[i];
+                    for(i = 0;i < MAX_SENDS;i++)
+                        ALSource->WetGains[i] = WetSend[i];
 
                 skipmix: ;
                 }
@@ -1338,9 +1231,11 @@ ALvoid aluMixData(ALCcontext *ALContext,ALvoid *buffer,ALsizei size,ALenum forma
         // effect slot processing
         while(ALEffectSlot)
         {
-            if(ALEffectSlot->effect.type == AL_EFFECT_REVERB)
-                VerbProcess(ALEffectSlot->ReverbState, SamplesToDo, WetBuffer, DryBuffer);
+            if(ALEffectSlot->EffectState)
+                ALEffect_Process(ALEffectSlot->EffectState, ALEffectSlot, SamplesToDo, ALEffectSlot->WetBuffer, DryBuffer);
 
+            for(i = 0;i < SamplesToDo;i++)
+                ALEffectSlot->WetBuffer[i] = 0.0f;
             ALEffectSlot = ALEffectSlot->next;
         }
 
